import os.path
from pathlib import Path

from safety.console import main_console as console
from safety.tool.utils import PipConfigurator, PipRequirementsConfigurator, PoetryPyprojectConfigurator, is_os_supported

from .interceptors import create_interceptor


def has_local_tool_files(directory: Path) -> bool:
    configurators = [PipRequirementsConfigurator(), PoetryPyprojectConfigurator()]

    for file_name in os.listdir(directory):
        if os.path.isfile(file_name):
            file = Path(file_name)
            for configurator in configurators:
                if configurator.is_supported(file):
                    return True

    return False


def configure_system():
    configurators = [PipConfigurator()]

    for configurator in configurators:
        configurator.configure()

def reset_system():
    configurators = [PipConfigurator()]

    for configurator in configurators:
        configurator.reset()

def configure_alias():
    if not is_os_supported():
        return    
    
    interceptor = create_interceptor()
    interceptor.install_interceptors()
    
    console.print("Configured PIP alias")


def configure_local_directory(directory: Path):
    configurators = [PipRequirementsConfigurator(), PoetryPyprojectConfigurator()]

    for file_name in os.listdir(directory):
        if os.path.isfile(file_name):
            file = Path(file_name)
            for configurator in configurators:
                if configurator.is_supported(file):
                    configurator.configure(file)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from safety.auth.models import Auth
    from rich.console import Console

    from safety_schemas.models import MetadataModel, ReportSchemaVersion, \
      TelemetryModel, PolicyFileModel, ConfigModel

@dataclass
class SafetyCLI:
    """
    A class representing Safety CLI settings.
    """
    auth: Optional['Auth'] = None
    telemetry: Optional['TelemetryModel'] = None
    metadata: Optional['MetadataModel'] = None
    schema: Optional['ReportSchemaVersion'] = None
    project = None
    config: Optional['ConfigModel'] = None
    console: Optional['Console'] = None
    system_scan_policy: Optional['PolicyFileModel'] = None
    platform_enabled: bool = False
    firewall_enabled: bool = False



import logging
import sys

from rich.prompt import Prompt
# TODO: refactor this import and the related code
# For now, let's keep it as is
from safety.scan.constants import DEFAULT_EPILOG
from ..cli_util import CommandType, FeatureType, SafetyCLICommand, \
    SafetyCLISubGroup, handle_cmd_exception, pass_safety_cli_obj
import typer


from safety.console import main_console as console

from ..constants import CONTEXT_COMMAND_TYPE, CONTEXT_FEATURE_TYPE, EXIT_CODE_OK

from .constants import FIREWALL_HELP, MSG_FEEDBACK, MSG_REQ_FILE_LINE, \
                       MSG_UNINSTALL_EXPLANATION, MSG_UNINSTALL_SUCCESS, \
                       UNINSTALL_CMD_NAME, UNINSTALL_HELP, FIREWALL_CMD_NAME, \
                       MSG_UNINSTALL_PIP_CONFIG, MSG_UNINSTALL_PIP_ALIAS

from ..tool.main import reset_system
from ..tool.interceptors import create_interceptor

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

firewall_app = typer.Typer(rich_markup_mode= "rich", cls=SafetyCLISubGroup, 
                           name=FIREWALL_CMD_NAME)


LOG = logging.getLogger(__name__)


init_app = typer.Typer(rich_markup_mode= "rich", cls=SafetyCLISubGroup)

@firewall_app.callback(cls=SafetyCLISubGroup,
                       help=FIREWALL_HELP,
                       epilog=DEFAULT_EPILOG,
                       context_settings={
                           "allow_extra_args": True,
                           "ignore_unknown_options": True,
                           CONTEXT_COMMAND_TYPE: CommandType.BETA,
                           CONTEXT_FEATURE_TYPE: FeatureType.FIREWALL
                       })
@pass_safety_cli_obj
def firewall(ctx: typer.Context) -> None:
    """
    Main callback for the firewall commands.

    Args:
        ctx (typer.Context): The Typer context object.
    """
    LOG.info('firewall callback started')


@firewall_app.command(
        cls=SafetyCLICommand,
        name=UNINSTALL_CMD_NAME,        
        help=UNINSTALL_HELP,
        options_metavar="[OPTIONS]",
        context_settings={
            "allow_extra_args": True,
            "ignore_unknown_options": True,
            CONTEXT_COMMAND_TYPE: CommandType.BETA,
            CONTEXT_FEATURE_TYPE: FeatureType.FIREWALL
        },
)
@handle_cmd_exception
def uninstall(ctx: typer.Context):
    console.print()
    console.print(MSG_UNINSTALL_EXPLANATION)
    
    console.print()
    prompt = "Uninstall?"
    should_uninstall = Prompt.ask(prompt=prompt, choices=["y", "n"], 
                                  default="y", show_default=True,
                                  console=console).lower() == 'y'

    if not should_uninstall:
        sys.exit(EXIT_CODE_OK)

    console.print()
    console.print(MSG_UNINSTALL_PIP_CONFIG)
    # TODO: Make it robust. The reset per tool should be included in remove
    # interceptors
    reset_system()

    # TODO: support reset project files

    console.print(MSG_UNINSTALL_PIP_ALIAS)
    interceptor = create_interceptor()
    interceptor.remove_interceptors()

    console.print()
    console.print(MSG_UNINSTALL_SUCCESS)
    
    console.print()
    console.print(MSG_REQ_FILE_LINE)

    console.print()

    # TODO: Ask for feedback
    # console.print(MSG_FEEDBACK)

    # console.print()
    # prompt = "Feedback (or enter to exit)"    
    # feedback = Prompt.ask(prompt)

    # if feedback:
    #     console.print()
    #     # TODO: send feedback to the server
    #     console.print("Thank you for your feedback!")





import abc
import os.path
import re
import subprocess
from abc import abstractmethod
from pathlib import Path
from sys import platform
from tempfile import mkstemp

import typer
from Levenshtein import distance
from filelock import FileLock
from rich.padding import Padding
from rich.prompt import Prompt

from safety.console import main_console as console
from safety.constants import PIP_LOCK
from safety.tool.constants import MOST_FREQUENTLY_DOWNLOADED_PYPI_PACKAGES, PROJECT_CONFIG, REPOSITORY_URL
from safety.tool.pip import Pip
from safety.tool.poetry import Poetry
from safety.tool.resolver import get_unwrapped_command

from typing_extensions import List

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from subprocess import CompletedProcess


def is_os_supported():
    return platform in ["linux", "linux2", "darwin", "win32"]

class BuildFileConfigurator(abc.ABC):

    @abc.abstractmethod
    def is_supported(self, file: Path) -> bool:
        """
        Returns whether a specific file is supported by this class.
        Args:
            file (str): The file to check.
        Returns:
            bool: Whether the file is supported by this class.
        """
        pass

    @abc.abstractmethod
    def configure(self, file: Path) -> None:
        """
        Configures specific file.
        Args:
            file (str): The file to configure.
        """
        pass


class PipRequirementsConfigurator(BuildFileConfigurator):
    __file_name_pattern = re.compile("^([a-zA-Z_-]+)?requirements([a-zA-Z_-]+)?.txt$")

    def is_supported(self, file: Path) -> bool:
        return self.__file_name_pattern.match(os.path.basename(file)) is not None

    def configure(self, file: Path) -> None:
        Pip.configure_requirements(file)


class PoetryPyprojectConfigurator(BuildFileConfigurator):
    __file_name_pattern = re.compile("^pyproject.toml$")

    def is_supported(self, file: Path) -> bool:
        return self.__file_name_pattern.match(os.path.basename(file)) is not None and Poetry.is_poetry_project_file(
            file)

    def configure(self, file: Path) -> None:
        Poetry.configure_pyproject(file)


# TODO: Review if we should move this/hook up this into interceptors.
class ToolConfigurator(abc.ABC):

    @abc.abstractmethod
    def configure(self) -> None:
        """
        Configures specific tool.
        """
        pass

    @abc.abstractmethod
    def reset(self) -> None:
        """
        Resets specific tool.
        """
        pass

class PipConfigurator(ToolConfigurator):

    def configure(self) -> None:
        Pip.configure_system()

    def reset(self) -> None:
        Pip.reset_system()


class PipCommand(abc.ABC):

    def __init__(self, args: List[str], capture_output: bool = False) -> None:
        self._args = args
        self.__capture_output = capture_output
        self.__filelock = FileLock(PIP_LOCK, 10)

    @abstractmethod
    def before(self, ctx: typer.Context):
        pass

    @abstractmethod
    def after(self, ctx: typer.Context, result):
        pass

    def execute(self, ctx: typer.Context):
        with self.__filelock:
            self.before(ctx)
            # TODO: Safety should redirect to the proper pip, if the user is
            # using pip3, it should be redirected to pip3, not pip to avoid any
            # issues.
            args = [get_unwrapped_command(name="pip")] + self.__remove_safety_args(self._args)
            result = subprocess.run(args, capture_output=self.__capture_output, env=self.env(ctx))
            self.after(ctx, result)

    def env(self, ctx: typer.Context):
        return os.environ.copy()

    @classmethod
    def from_args(self, args):
        if "install" in args:
            return PipInstallCommand(args)
        elif "uninstall" in args:
            return PipUninstallCommand(args)
        else:
            return PipGenericCommand(args)

    def __remove_safety_args(self, args: List[str]):
        return [arg for arg in args if not arg.startswith("--safety")]


class PipGenericCommand(PipCommand):

    def __init__(self, args: List[str]) -> None:
        super().__init__(args)

    def before(self, ctx: typer.Context):
        pass

    def after(self, ctx: typer.Context, result):
        pass


class PipInstallCommand(PipCommand):

    def __init__(self, args: List[str]) -> None:
        super().__init__(args)
        self.package_names = []
        self.__index_url = None

    def before(self, ctx: typer.Context):
        args = self._args

        ranges_to_delete = []
        for ind, val in enumerate(args):
            if ind > 0 and (args[ind - 1].startswith("-i") or args[ind - 1].startswith("--index-url")):
                if args[ind].startswith("https://pkgs.safetycli.com"):
                    self.__index_url = args[ind]

                ranges_to_delete.append((ind - 1, ind))
            elif ind > 0 and (args[ind - 1] == "-r" or args[ind - 1] == "--requirement"):
                requirement_file = args[ind]

                if not Path(requirement_file).is_file():
                    continue

                with open(requirement_file, "r") as f:
                    fd, tmp_requirements_path = mkstemp(suffix="safety-requirements.txt", text=True)
                    with os.fdopen(fd, "w") as tf:
                        requirements = re.sub(r"^(-i|--index-url).*$", "", f.read(), flags=re.MULTILINE)
                        tf.write(requirements)

                    args[ind] = tmp_requirements_path
            elif ind > 0 and (not args[ind - 1].startswith("-e") or not args[ind - 1].startswith("--editable")) and not args[ind].startswith("-"):
                if args[ind] == '.':
                    continue

                package_name = args[ind]
                (valid, candidate_package_name) = self.__check_typosquatting(package_name)
                if not valid:
                    prompt = f"You are about to install {package_name} package. Did you mean to install {candidate_package_name}?"
                    answer = Prompt.ask(prompt=prompt, choices=["y", "n"],
                                        default="y", show_default=True, console=console).lower()
                    if answer == 'y':
                        package_name = candidate_package_name
                        console.print(f"Installing {package_name} package instead.")
                        args[ind] = package_name

                self.__add_package_name(package_name)

        for (start, end) in ranges_to_delete:
            args = args[:start] + args[end + 1:]

        self._args = args

    def after(self, ctx: typer.Context, result: 'CompletedProcess[str]'):
        if result and result.returncode == 0:
            self.__run_scan()
        else:
            self.__render_package_details()

    def env(self, ctx: typer.Context) -> dict:
        env = super().env(ctx)
        env["PIP_INDEX_URL"] = Pip.build_index_url(ctx, self.__index_url) if not self.__is_check_disabled() else Pip.default_index_url()
        return env

    def __is_check_disabled(self):
        return "--safety-disable-check" in self._args

    def __check_typosquatting(self, package_name):
        max_edit_distance = 2 if len(package_name) > 5 else 1

        if package_name in MOST_FREQUENTLY_DOWNLOADED_PYPI_PACKAGES:
            return (True, package_name)

        for pkg in MOST_FREQUENTLY_DOWNLOADED_PYPI_PACKAGES:
            if (abs(len(pkg) - len(package_name)) <= max_edit_distance
                and distance(pkg, package_name) <= max_edit_distance):
                return (False, pkg)

        return (True, package_name)

    def __run_scan(self):
        if not is_os_supported():
            return

        target = os.getcwd()
        if Path(os.path.join(target, PROJECT_CONFIG)).is_file():
            try:
                subprocess.Popen(
                ['safety', 'scan'],
                     stdout=subprocess.DEVNULL,
                     stderr=subprocess.DEVNULL,
                     stdin=subprocess.DEVNULL,
                     start_new_session=True
                )
            except Exception:
                pass

    def __add_package_name(self, package_name):
        r = re.compile(r"^([a-zA-Z_-]+)(([~<>=]=)[a-zA-Z0-9._-]+)?")
        match = r.match(package_name)
        if match:
            self.package_names.append(match.group(1))

    def __render_package_details(self):
        for package_name in self.package_names:
            console.print(
                Padding(f"Learn more: [link]https://data.safetycli.com/packages/pypi/{package_name}/[/link]",
                        (0, 0, 0, 1)), emoji=True)


class PipUninstallCommand(PipCommand):

    def __init__(self, args: List[str]) -> None:
        super().__init__(args)

    def before(self, ctx: typer.Context):
        pass

    def after(self, ctx: typer.Context, result):
        pass



from pathlib import Path

from rich.prompt import Prompt
from ..cli_util import CommandType, FeatureType, SafetyCLICommand, \
    SafetyCLISubGroup, handle_cmd_exception
import typer
import os


from safety.init.constants import PROJECT_INIT_CMD_NAME, PROJECT_INIT_HELP, PROJECT_INIT_DIRECTORY_HELP
from safety.init.main import create_project
from safety.console import main_console as console
from ..scan.command import scan
from ..scan.models import ScanOutput
from ..tool.main import configure_system, configure_local_directory, has_local_tool_files, configure_alias

from ..constants import CONTEXT_COMMAND_TYPE, CONTEXT_FEATURE_TYPE

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

init_app = typer.Typer(rich_markup_mode= "rich", cls=SafetyCLISubGroup)

@init_app.command(
        cls=SafetyCLICommand,
        help=PROJECT_INIT_HELP,
        name=PROJECT_INIT_CMD_NAME,
        options_metavar="[OPTIONS]",
        context_settings={
            "allow_extra_args": True,
            "ignore_unknown_options": True,
            CONTEXT_COMMAND_TYPE: CommandType.BETA,
            CONTEXT_FEATURE_TYPE: FeatureType.FIREWALL
        },
)
@handle_cmd_exception
def init(ctx: typer.Context,
         directory: Annotated[
             Path,
             typer.Argument(
                 exists=True,
                 file_okay=False,
                 dir_okay=True,
                 writable=False,
                 readable=True,
                 resolve_path=True,
                 show_default=False,
                 help=PROJECT_INIT_DIRECTORY_HELP
             ),
         ] = Path(".")):

    do_init(ctx, directory, False)


def do_init(ctx: typer.Context, directory: Path, prompt_user: bool = True):
    project_dir = directory if os.path.isabs(directory) else os.path.join(os.getcwd(), directory)
    create_project(ctx, console, Path(project_dir))

    answer = 'y' if not prompt_user else None
    if prompt_user:
        console.print(
            "Safety prevents vulnerable or malicious packages from being installed on your computer. We do this by wrapping your package manager.")
        prompt = "Do you want to enable proactive malicious package prevention?"
        answer = Prompt.ask(prompt=prompt, choices=["y", "n"],
                            default="y", show_default=True, console=console).lower()

    if answer == 'y':
        configure_system()

    if prompt_user:
        prompt = "Do you want to alias pip to Safety?"
        answer = Prompt.ask(prompt=prompt, choices=["y", "n"],
                            default="y", show_default=True, console=console).lower()

    if answer == 'y':
        configure_alias()

    if has_local_tool_files(project_dir):
        if prompt_user:
            prompt = "Do you want to enable proactive malicious package prevention for any project in working directory?"
            answer = Prompt.ask(prompt=prompt, choices=["y", "n"],
                                default="y", show_default=True, console=console).lower()

        if answer == 'y':
            configure_local_directory(project_dir)

        if prompt_user:
            prompt = "It looks like your current directory contains a requirements.txt file. Would you like Safety to scan it?"
            answer = Prompt.ask(prompt=prompt, choices=["y", "n"],
                                default="y", show_default=True, console=console).lower()
        
        if answer == 'y':
            ctx.command.name = "scan"
            ctx.params = {
                "target": directory,
                "output": ScanOutput.SCREEN,
                "policy_file_path": None
            }
            scan(ctx=ctx, target=directory, output=ScanOutput.SCREEN, policy_file_path=None)



from sys import platform
from typing import Optional
from .types import InterceptorType
from .unix import UnixAliasInterceptor
from .windows import WindowsInterceptor
from .base import CommandInterceptor


def create_interceptor(interceptor_type: 
                       Optional[InterceptorType] = None) -> CommandInterceptor:
    """
    Create appropriate interceptor based on OS and type
    """
    interceptor_map = {
        InterceptorType.UNIX_ALIAS: UnixAliasInterceptor,
        InterceptorType.WINDOWS_BAT: WindowsInterceptor
    }
    
    if interceptor_type:
        return interceptor_map[interceptor_type]()
        
    # Auto-select based on OS
    if platform == 'win32':
        return interceptor_map[InterceptorType.WINDOWS_BAT]()
    
    if platform in ['linux', 'linux2', 'darwin']:
        # Default to alias-based on Unix-like systems
        return interceptor_map[InterceptorType.UNIX_ALIAS]()
    
    raise NotImplementedError(f"Platform '{platform}' is not supported.")



from .obj import SafetyCLI
from .requirements import is_pinned_requirement
from .vulnerabilities import Vulnerability, CVE, Severity, Fix, \
    SafetyRequirement, Package, SafetyEncoder, RequirementFile


__all__ = [
    'Package',
    'SafetyCLI',
    'Vulnerability',
    'CVE', 
    'Severity',
    'Fix',
    'is_pinned_requirement',
    'SafetyRequirement',
    'SafetyEncoder',
    'RequirementFile'
]


import shutil
import subprocess
from pathlib import Path
from typing import Optional
import sys

from rich.console import Console

from safety.console import main_console
from safety.tool.pip import REPOSITORY_URL
from safety.tool.resolver import get_unwrapped_command

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

class Poetry:

    @classmethod
    def is_installed(cls) -> bool:
        """
        Checks if the PIP program is installed

        Returns:
            True if PIP is installed on system, or false otherwise
        """
        return shutil.which("poetry") is not None

    @classmethod
    def is_poetry_project_file(cls, file: Path) -> bool:
        try:
            cfg = tomllib.loads(file.read_text())
            return cfg.get("build-system", {}).get("requires") == "poetry-core"
        except (IOError, ValueError) as e:
            return False

    @classmethod
    def configure_pyproject(cls, file: Path, console: Optional[Console] = main_console) -> None:
        """
        Configures index url for specified requirements file.

        Args:
            file (Path): Path to requirements.txt file.
            console (Console): Console instance.
        """
        if not cls.is_installed():
            console.log("Poetry is not installed.")

        subprocess.run([get_unwrapped_command(name="poetry"), "source", "add", "safety", REPOSITORY_URL], capture_output=True)
        console.print(f"Configured {file} file")
